"""
Metrics for data synthesis
"""
import pandas as pd
import numpy as np
from sklearn.metrics import normalized_mutual_info_score, confusion_matrix


def pairwise_mutual_information(frame: pd.DataFrame):
    """
    Return mutual information matrix for pairwise columns of a DataFrame.
    """
    columns = frame.columns.sort_values().to_list()
    mi = pd.DataFrame(columns=columns, index=columns, dtype=float)
    for row in columns:
        for col in columns:
            if pd.isnull(mi.at[row, col]):
                mi.at[row, col] = normalized_mutual_info_score(
                    frame[row].astype(str), frame[col].astype(str),
                    average_method='arithmetic')
            else:
                mi.at[row, col] = mi.at[col, row]
    return mi.round(3)


def jensen_shannon_divergence(p, q, base=2):
    """
    Return the Jensen-Shannon divergence between two 1-D arrays.

    Parameters
    ---------
    p : array
        left probability array
    q : array
        right probability array
    base : numeric, default 2
        logarithm base

    Returns
    -------
    jsd : float
        divergence of p and q
    """
    # If the sum of probability array p or q does not equal to 1, then normalize
    p = np.asarray(p)
    q = np.asarray(q)
    p = p / np.sum(p, axis=0)
    q = q / np.sum(q, axis=0)
    from scipy.spatial.distance import jensenshannon
    return round(jensenshannon(p, q, base=base), 4)


def error_rate(y_true, y_pred=None):
    """
    Return error (mis-classification) rate of one classifier result

    If there is only one parameter, it must be the confusion matrix;
    If there are two parameters, they must be true and predict labels;
    """
    if y_pred is None:
        if isinstance(y_true, pd.DataFrame):
            cm = y_true.values
        else:
            cm = y_true
    else:
        cm = confusion_matrix(y_true, y_pred)
    trace = np.trace(cm)
    sum = np.sum(cm)
    return round((sum - trace) / sum, 4)


def relative_error(x, y):
    """
    Return relative error of two variables: |x-y|/max(|x|, |y|)
    """
    m = np.maximum(np.abs(x), np.abs(y))
    return round(np.average(np.abs(x - y) / (m + 1e-6)), 4)



# -*- coding: utf-8 -*-

"""
Utility functions for data synthesis. Including:
    input/output,
    machine learning,
    ...
"""
import argparse
import csv
import logging
import numpy as np
import os

from string import ascii_lowercase
from pandas import Series, DataFrame

logger = logging.getLogger(__name__)


# ---------------------------------------------------------------
# Utilities for Input/Output


def has_header(path, encoding='utf-8'):
    """
    Auto-detect if csv file has header.
    """
    from pandas import read_csv
    df0 = read_csv(path, header=None, nrows=10, skipinitialspace=True,
                   encoding=encoding)
    df1 = read_csv(path, nrows=10, skipinitialspace=True, encoding=encoding)
    # If the column is numerical, its dtype is different without/with header
    # TODO how about categorical columns
    return tuple(df0.dtypes) != tuple(df1.dtypes)


def read_data_from_csv(path, na_values=None, header=None):
    """
    Read data set from csv or other delimited text file. And remove empty
    columns (all values are null).
    """
    from pandas import read_csv

    try:
        header = header or ('infer' if has_header(path) else None)
        df = read_csv(path, skipinitialspace=True, na_values=na_values,
                      header=header)
    except (UnicodeDecodeError, NameError):
        header = header or ('infer' if has_header(path, encoding='latin1')
                            else None)
        df = read_csv(path, skipinitialspace=True, na_values=na_values,
                      header=header, encoding='latin1')

    # Remove columns with empty active domain, i.e., all values are missing.
    before_attrs = set(df.columns)
    df.dropna(axis=1, how='all')
    after_attrs = set(df.columns)
    if len(before_attrs) > len(after_attrs):
        print(
            f'Empty columns are removed, include {before_attrs - after_attrs}.')
    # Remove rows with all empty values
    df.dropna(axis=0, how='all')
    if header is None:
        df = df.rename(lambda x: f'#{x}', axis='columns')
    return df


def write_csv(file, data):
    """ write data to csv files """
    # if data is not a list of list, make it:
    if isinstance(data, list) and not isinstance(data[0], list):
        data = [data]
    with open(file, 'a') as f:
        writer = csv.writer(f, lineterminator='\n')
        writer.writerows(data)
    f.close()


def file_name(path):
    """ Return the file name without extension from a path. """
    return os.path.splitext(os.path.basename(path))[0]


def str_to_list(val, separator=','):
    """
    Split one string to a list by separator.
    """
    if val is None:
        return None
    return val.split(separator)


# ---------------------------------------------------------------
# Utilities for Plotting


def _compress_svg_data(svg: str):
    import re
    value = re.sub(r'\n', ' ', svg)
    value = re.sub(r' {2,}', ' ', value)
    value = re.sub(r'<style type="text/css">.*</style>', '', value)
    value = re.sub(r'<!--(.*?)-->', '', value)
    value = re.sub(r'<\?xml.*\?>', '', value)
    value = re.sub(r'<!DOCTYPE.*dtd">', '', value)
    return value.strip()


def _prepare_for_cjk_characters(chars):
    """ If input string has Chinese, Japanese, and Korean characters, set
    specific font for them. """
    has_cjk = False
    for c in chars:
        if any([start <= ord(c) <= end for start, end in [
            (4352, 4607), (11904, 42191), (43072, 43135), (44032, 55215),
            (63744, 64255), (65072, 65103), (65381, 65500), (131072, 196607)
        ]]):
            has_cjk = True
            break
    if has_cjk:
        import matplotlib
        matplotlib.rcParams['font.family'] = ['Microsoft Yahei']


def plot_confusion_matrix(matrix: DataFrame, title='',
                          ylabel='Predicted', xlabel='Actual',
                          otype='string', path=None, cmap='Blues',
                          vrange=None):
    """
    Plot a confusion matrix to show predict and actual values.

    Parameters
    ----------
    matrix : pandas.DataFrame
        the matrix to plot

    title : str
        title of image

    ylabel : str
        label in y-axis of image

    xlabel : str
        label in x-axis of image

    otype : str
        output type, support 'string' (default), 'file', 'show'

    path : str
        output path for 'file' type, default is 'matrix_2_3.svg' ((2, 3) is the
        shape of matrix).

    cmap : str
        color

    vrange : 2-tuple
        manually set range of matrix
    """
    import matplotlib.pyplot as plt
    _prepare_for_cjk_characters(''.join(matrix.columns))
    figsize = (6.4, 4.8)
    n_columns = matrix.columns.size
    if n_columns > 9:
        from math import ceil
        width = 6.4 + ceil((n_columns - 9) / 2) * 0.8
        height = width * figsize[1] / figsize[0]
        figsize = (width, height)
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    if vrange is None:
        vrange = (matrix.values.min(), matrix.values.max())
    exp = len(str(vrange[1])) - 1
    if exp > 2:
        matrix = matrix.div(10 ** exp)
        vrange = (vrange[0] / (10 ** exp), vrange[1] / (10 ** exp))
    im = ax.imshow(matrix.values, cmap=cmap, vmin=vrange[0], vmax=vrange[1])
    ax.set(xticks=np.arange(matrix.shape[1]),
           yticks=np.arange(matrix.shape[0]),
           xticklabels=matrix.columns,
           yticklabels=matrix.index,
           title=title,
           ylabel=ylabel,
           xlabel=xlabel)
    ax.tick_params(which='both', length=0)
    for edge, spine in ax.spines.items():
        spine.set_visible(False)

    # Loop over data dimensions and create text annotations.
    thresh = (vrange[0] + vrange[1]) / 2.
    for i in range(matrix.shape[0]):
        for j in range(matrix.shape[1]):
            if exp > 2:
                text = '{:.2f}'.format(matrix.iloc[i, j])
            else:
                text = matrix.iloc[i, j]
            ax.text(j, i, text,
                    ha="center", va="center",
                    color="white" if matrix.iloc[i, j] > thresh else "black")
    from ds4mllib.metrics import error_rate
    ax.set_title('Rate: {:.1%}'.format(error_rate(matrix)), fontsize=10)

    cbar = ax.figure.colorbar(im, ax=ax, pad=0.03, aspect=30)
    cbar.outline.set_visible(False)
    if exp > 2:
        cbar.ax.text(0, -0.1, f'(e+{exp})')
    cbar.ax.tick_params(which='both', length=0)
    fig.autofmt_xdate()
    fig.tight_layout()
    plt.subplots_adjust()
    try:
        if otype == 'string':
            from io import StringIO
            img = StringIO()
            plt.savefig(img, format='svg', bbox_inches='tight')
            return _compress_svg_data(img.getvalue())
        elif otype == 'file':
            if path is None:
                path = 'matrix_{}_{}.svg'.format(matrix.shape[0],
                                                 matrix.shape[1])
            plt.savefig(path, bbox_inches='tight')
            return path
        elif otype == 'show':
            plt.show()
    finally:
        plt.close()


def plot_histogram(bins, heights, otype='string', path=None,
                   x_label='', y_label='Frequency'):
    """
    Plot two bars.
    Note: Because pyplot draws diagrams auto-scale, this function will provide
    kinds of diagram patterns.
    """
    import matplotlib.pyplot as plt
    _prepare_for_cjk_characters(''.join(map(str, bins)))
    length = len(bins)
    fig_width = 6.4
    ticks = 22
    if length < 5:
        # make 7 bars to show the data
        # insert 0 or '' to the center of array
        # e.g. [3, 4] => [3, 0, 4], ['1', '2', '3'] => ['1', '', '2', '', '3'], ...
        bins = list(map(str, bins))
        bins = ',,'.join(bins).split(',')
        heights = np.insert(heights, list(range(1, length)), 0, axis=1)
        # pad 0 or '' to array in the begin and end
        pad = (7 - len(bins)) // 2
        bins = tuple([''] * pad + bins + [''] * pad)
        heights = np.append(np.insert(heights, [0], [0] * pad, axis=1),
                            np.zeros((len(heights), pad), dtype=int), axis=1)
        length = 7
    else:
        # TODO: if count of bins is bigger than 33, and it is categorical, how?
        if length >= 60:
            bins = bins[:60]
            heights = heights[:, :60]
            length = 60
        bins = tuple(map(str, bins))
        length_ = [8, 12, 16, 20, 32, 42, 48, 60]
        ticks_ = [22, 30, 36, 48, 76, 96, 108, 172]
        width_ = [6.4, 9.6, 11.2, 11.2, 11.2, 14, 14, 15]
        idx = len([i for i in length_ if length > i])
        ticks = ticks_[idx]
        fig_width = width_[idx]

    x = np.arange(length)
    fig = plt.figure(figsize=(fig_width, 4.8))
    ax = fig.add_subplot(111)
    width = length / ticks
    diff = width / 2 + 0.01
    ax.bar(x - diff, heights[0], width=width, color='#38bbe8')
    ax.bar(x + diff, heights[1], width=width, color='#ffd56e')
    ax.legend(['raw', 'synth'])
    fig.autofmt_xdate()
    # fig.tight_layout(pad=1.5)
    plt.xticks(x, bins, fontsize=10)
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.subplots_adjust()
    try:
        if otype == 'string':
            from io import StringIO
            img = StringIO()
            plt.savefig(img, format='svg', bbox_inches='tight')
            return _compress_svg_data(img.getvalue())
        elif otype == 'file':
            if path is None:
                path = 'histogram_{}_{}.svg'.format(len(bins), len(heights))
            plt.savefig(path, bbox_inches='tight')
            return path
        elif otype == 'show':
            plt.show()
    finally:
        plt.close()


def plot_heatmap(data, title='', otype='string', path=None, cmap='Blues'):
    """
    Plot a heatmap to show pairwise correlations (e.g. mutual information).

    Parameters see <code>plot_confusion_matrix</code>
    """
    import matplotlib.pyplot as plt
    _prepare_for_cjk_characters(''.join(data.columns))
    fig = plt.figure()
    ax = fig.add_subplot(111)
    im = ax.imshow(data.values, cmap=cmap)

    ax.set_title(title, fontsize=10)
    ticks = np.arange(len(data.columns))
    ax.set_xticks(ticks)
    ax.set_xticklabels(data.columns)
    ax.set_yticks(ticks)
    ax.set_yticklabels(data.index)
    ax.tick_params(which='both', length=0)
    for edge, spine in ax.spines.items():
        spine.set_visible(False)

    # set color bar in the right
    cbar = ax.figure.colorbar(im, ax=ax, pad=0.03, aspect=30)
    cbar.outline.set_visible(False)
    cbar.ax.tick_params(which='both', length=0)
    fig.autofmt_xdate()
    fig.tight_layout()
    plt.subplots_adjust()
    try:
        if otype == 'string':
            from io import StringIO
            img = StringIO()
            plt.savefig(img, format='svg', facecolor='#ebebeb',
                        bbox_inches='tight')
            return _compress_svg_data(img.getvalue())
        elif otype == 'file':
            if path is None:
                path = 'heatmap_{}_{}.svg'.format(data.shape[0], data.shape[1])
            plt.savefig(path, facecolor='#ebebeb', bbox_inches='tight')
            return path
        elif otype == 'show':
            plt.show()
    finally:
        plt.close()


# ---------------------------------------------------------------
# Utilities for Metrics, ML

def train_and_predict(x_train, y_train, x_test):
    """
    Predict <x, y> by SVM classifier, and compare with test data
    TODO: do some analysis and then choose svm classifier and its parameters
    """
    from sklearn.svm import SVC
    classifier = SVC(gamma='scale')
    classifier.fit(x_train, y_train)
    result = classifier.predict(x_test)
    return result


def mutual_information(child: Series, parents: DataFrame):
    """
    Mutual information of child (Series) and parents (DataFrame) distributions
    """
    from sklearn.metrics import mutual_info_score
    if parents.shape[1] == 1:
        parents = parents.iloc[:, 0]
    else:
        parents = parents.apply(lambda x: ' '.join(x.get_values()), axis=1)
    return mutual_info_score(child, parents)


def normalize_distribution(frequencies):
    frequencies = np.array(frequencies, dtype=float)
    frequencies = frequencies.clip(0)
    total = frequencies.sum()
    if total > 0:
        if np.isinf(total):
            return normalize_distribution(np.isinf(frequencies))
        else:
            return frequencies / total
    else:
        return np.full_like(frequencies, 1 / frequencies.size)


def normalize_range(start, stop, bins=20):
    """
    Return evenly spaced values within a given interval, and a dynamically
    calculated step, to make number of bins close to 20. If integer interval,
    the smallest step is 1; If float interval, the smallest step is 0.5.
    """
    from math import ceil, floor
    if isinstance(start, int) and isinstance(stop, int):
        step = ceil((stop - start) / bins)
    else:
        start = floor(start)
        stop = ceil(stop)
        step = (stop - start) / bins
        step = ceil(step) if ceil(step) == round(step) else round(step) + 0.5
    stop = step * (bins + 1) + start
    return np.arange(start, stop, step)


def is_datetime(value: str):
    from dateutil.parser import parse
    """
    Detect a value is a datetime. Exclude some datetime literals (weekdays and
    months) from method `dateutil.parser`.
    """
    literals = {'mon', 'monday', 'tue', 'tuesday', 'wed', 'wednesday', 'thu',
                'thursday', 'fri', 'friday', 'sat', 'saturday', 'sun', 'sunday',
                'jan', 'january', 'feb', 'february', 'mar', 'march', 'apr',
                'april', 'may', 'jun', 'june', 'jul', 'july', 'aug', 'august',
                'sep', 'sept', 'september', 'oct', 'october', 'nov', 'november',
                'dec', 'december'}
    try:
        value = value.lower()
        if value in literals:
            return False
        parse(value)
        return True
    except ValueError:
        return False
    except AttributeError:
        return False


def randomize_string(length):
    return ''.join(np.random.choice(list(ascii_lowercase), size=length))


def pseudonymise_string(value):
    chars = 'oVT0X5Is#94OpnDifAkRGaUP3KcNYhjzd8CWmQurybBgqMZH7FL2tJe$vxEw1Sl6'
    return ''.join(map(lambda a: chars[ord(a) & 63], list(value)))


# ---------------------------------------------------------------
# Utilities for arguments parser


class CustomFormatter(argparse.HelpFormatter):
    def _format_action_invocation(self, action):
        if not action.option_strings:
            metavar, = self._metavar_formatter(action, action.dest)(1)
            return metavar
        else:
            parts = []
            # if the Optional doesn't take a value, format is:
            #    -s, --long
            if action.nargs == 0:
                parts.extend(action.option_strings)

            # if the Optional takes a value, format is:
            #    -s ARGS, --long ARGS
            # change to
            #    -s, --long ARGS
            else:
                default = action.dest.upper()
                args_string = self._format_args(action, default)
                for option_string in action.option_strings:
                    # parts.append('%s %s' % (option_string, args_string))
                    parts.append('%s' % option_string)
                parts[-1] += ' %s' % args_string
            return ', '.join(parts)



# encoding: utf-8
"""
Algorithms to synthesize data set: currently differential privacy
"""
import logging
import warnings
import numpy as np

from itertools import product
from pandas import DataFrame, merge
from scipy.optimize import fsolve

from ds4mllib.utils import mutual_information, normalize_distribution

logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
# Algorithms PrivBayes: Private Data Release via Bayesian Networks


def calculate_sensitivity(n_rows, child, parents, binaries):
    """
    Lemma 4.1 Page 12: Sensitivity function for Bayesian network construction.

    Parameters
    ----------
    n_rows : int
        Number of tuples in dataset
    child : str
        One column name
    parents : tuple
        Parents of child, there may be multiple parents
    binaries : list
        List of binary columns
    """
    if child in binaries or (len(parents) == 1 and parents[0] in binaries):
        a = np.log(n_rows) / n_rows
        b = (1 / n_rows - 1) * np.log(1 - 1 / n_rows)
    else:
        a = (2 / n_rows) * np.log((n_rows + 1) / 2)
        b = (1 - 1 / n_rows) * np.log(1 + 2 / (n_rows - 1))
    return a + b


def usefulness(degree, n_rows, n_cols, epsilon, threshold):
    """
    Lemma 4.8 Page 16: Usefulness measure of each noisy marginal distribution
    """
    theta = n_rows * epsilon / ((n_cols - degree) * (2 ** (degree + 2)))
    return theta - threshold


def calculate_degree(n_rows, n_cols, epsilon):
    """
    Lemma 5.8 Page 16: The largest degree that guarantee theta-usefulness.
    """
    threshold = 5  # threshold of usefulness from Page 17
    default = min(3, int(n_cols / 2))
    args = (n_rows, n_cols, epsilon, threshold)
    warnings.filterwarnings("error")
    try:
        degree = fsolve(usefulness, np.array(int(n_cols / 2)), args=args)[0]
        degree = int(np.ceil(degree))
    except RuntimeWarning:
        warnings.warn('Degree of bayesian network is not properly computed!')
        degree = default
    if degree < 1 or degree > n_cols:
        degree = default
    return degree


def candidate_pairs(paras):
    """
    Return attribute-parents pairs, and their mutual information.
    """
    from itertools import combinations
    child, columns, n_parents, index, dataset = paras
    aps = []
    mis = []

    if index + n_parents - 1 < len(columns):
        for parents in combinations(columns[index + 1:], n_parents - 1):
            parents = list(parents)
            parents.append(columns[index])
            aps.append((child, parents))
            # TODO duplicate calculation of mutual information
            mi = mutual_information(dataset[child], dataset[parents])
            mis.append(mi)
    return aps, mis


def greedy_bayes(dataset: DataFrame, epsilon, degree=None, retains=None):
    """
    Algorithm 4, Page 20: Construct bayesian network by greedy algorithm.

    Parameters
    ----------
    dataset : DataFrame
        Encoded dataset
    epsilon : float
        Parameter of differential privacy
    degree : int
        Degree of bayesian network. If null, calculate it automatically.
    retains : list
        The columns to retain
    """
    dataset = dataset.astype(str, copy=False)
    n_rows, n_cols = dataset.shape
    retains = retains or []
    if not degree:
        degree = calculate_degree(n_rows, n_cols, epsilon)

    # mapping from column name to is_binary, because sensitivity is different
    # for binary or non-binary column
    binaries = [col for col in dataset if dataset[col].unique().size > 2]
    more_retains = False
    if len(retains) == 0:
        root_col = np.random.choice(dataset.columns)
    elif len(retains) == 1:
        root_col = retains[0]
    else:
        root_col = np.random.choice(retains)
        more_retains = True

    # columns: a set that contains all attributes whose parent sets has been set
    columns = [root_col]
    if more_retains:
        left_cols = set(retains)
    else:
        left_cols = set(dataset.columns)
    left_cols.remove(root_col)
    network = []
    while len(left_cols) > 0:
        # ap: attribute-parent (AP) pair is a tuple. It is a node in bayesian
        # network, e.g. ('education', ['relationship']), there may be multiple
        # parents, depends on k (degree of bayesian network).
        aps = []
        # mi: mutual information (MI) of two features
        mis = []
        n_parents = min(len(columns), degree)
        # calculate the candidate set of attribute-parent pair
        tasks = [(child, columns, n_parents, index, dataset) for child, index in
                 product(left_cols, range(len(columns) - n_parents + 1))]
        # TODO: should use thread pool for large data set?
        candidates = list(map(candidate_pairs, tasks))
        for ap, mi in candidates:
            aps += ap
            mis += mi
        # find next child node in bayesian networks according to the biggest
        # mutual information or exponential mechanism
        if epsilon:
            index = sampling_pair(mis, aps, binaries, n_rows, n_cols, epsilon)
        else:
            index = mis.index(max(mis))
        network.append(aps[index])
        next_col = aps[index][0]
        columns.append(next_col)
        left_cols.remove(next_col)
        if len(left_cols) == 0 and more_retains:
            left_cols = set(dataset.columns) - set(retains)
            more_retains = False
    return network


def sampling_pair(mis, aps, binaries, n_rows, n_cols, epsilon):
    """
    Page 6 and 12: Sampling an attribute-parent pair from candidates by
    exponential mechanism.
    """
    deltas = []
    for child, parents in aps:
        sensitivity = calculate_sensitivity(n_rows, child, parents, binaries)
        delta = (n_cols - 1) * sensitivity / epsilon
        deltas.append(delta)
    prs = np.array(mis) / (2 * np.array(deltas))
    prs = np.exp(prs)
    prs = normalize_distribution(prs)
    return np.random.choice(list(range(len(mis))), p=prs)


def noisy_distributions(dataset, columns, epsilon):
    """
    Calculate differentially private distribution
    """
    data = dataset.copy()[columns]
    data['freq'] = 1
    freq = data.groupby(columns).sum()
    freq.reset_index(inplace=True)

    iters = [range(int(dataset[col].max()) + 1) for col in columns]
    domain = DataFrame(columns=columns, data=list(product(*iters)))
    freq = merge(domain, freq, how='left')
    freq.fillna(0, inplace=True)

    n_rows, n_cols = dataset.shape
    scale = 2 * n_cols / (n_rows * epsilon)
    # if all noisy distributions are zero, add noises again
    while True:
        noises = np.random.laplace(0, scale=scale, size=freq.shape[0])
        freq['freq'] += noises
        freq.loc[freq['freq'] < 0, 'freq'] = 0
        if freq['freq'].sum() > 0:
            break
    return freq


def noisy_conditionals(network, dataset, epsilon):
    """
    Algorithm 3, Page 20: noisy conditional distribution probability
    """
    cond_prs = {}  # conditional probability distributions

    # distribution of one or more root node(s) in bayesian network
    for col in network[0][1]:
        freq = noisy_distributions(dataset, [col], epsilon)
        prs = freq[[col, 'freq']].groupby(col).sum()['freq']
        cond_prs[col] = normalize_distribution(prs).tolist()

    # distributions of other child node(s) in bayesian network
    for child, parents in network:
        cond_prs[child] = {}
        freq = noisy_distributions(dataset, parents + [child], epsilon)
        freq = DataFrame(freq[parents + [child, 'freq']]
                         .groupby(parents + [child]).sum())
        if len(parents) == 1:
            for parent in freq.index.levels[0]:
                prs = normalize_distribution(freq.loc[parent]['freq']).tolist()
                cond_prs[child][str([parent])] = prs
        else:
            for parent in product(*freq.index.levels[:-1]):
                prs = normalize_distribution(freq.loc[parent]['freq']).tolist()
                cond_prs[child][str(list(parent))] = prs
    return cond_prs



__version__ = '0.0.1'

from ds4mllib.dataset import DataSet
from ds4mllib.attribute import Attribute
from ds4mllib.evaluator import BiFrame

# Default environment variables for data processing and analysis
# It's not recommend to modify these variables' values.
params = {
    'attribute.bins': 20
}



"""
BiFrame for data synthesis.
"""

import logging
import numpy as np
import pandas as pd

from pandas import Index
from sklearn.model_selection import train_test_split

from ds4mllib.dataset import DataSet
from ds4mllib.utils import train_and_predict, normalize_range
from ds4mllib.metrics import jensen_shannon_divergence, relative_error

logger = logging.getLogger(__name__)


class BiFrame(object):
    def __init__(self, first: pd.DataFrame, second: pd.DataFrame):
        """
        BiFrame class that contains two data sets, which currently provides
        kinds of analysis methods from distribution, correlation, and some
        machine learning tasks.
        Especially, if the input data sets are source and synthesized dataset,
        this class can be used to evaluate the utility and privacy of
        synthesized data set.

        Parameters
        ----------
        first : {pandas.DataFrame}
            first data set

        second : {pandas.DataFrame}
            second data set, i.e. synthesized ata
        """
        # distribution
        self._dt = {}

        # To compare two data set, make sure that they have same columns.
        # If not, compare the common part.
        common = set(first.columns) & set(second.columns)
        if len(common) != len(first.columns) or len(common) != len(second.columns):
            logger.info(f"BiFrame constructed on attributes: {common}.")

        # left and right data set (ds)
        self.first = DataSet(first[common])
        self.second = DataSet(second[common])
        self._columns = self.first.columns.sort_values().to_list()

        # Make sure that two dataset have same domain for categorical
        # attributes, and same min, max values for numerical attributes.
        for col in self._columns:
            # If current column is not categorical, will ignore it.
            if not self.first[col].categorical or not self.second[col].categorical:
                continue
            d1, d2 = self.first[col].domain, self.second[col].domain
            if not np.array_equal(d1, d2):
                if self.first[col].categorical:
                    domain = np.unique(np.concatenate((d1, d2)))
                else:
                    domain = [min(d1[0], d2[0]), max(d1[1], d2[1])]
                self.first[col].domain = domain
                self.second[col].domain = domain

    @property
    def columns(self):
        return self._columns

    def err(self):
        """
        Return pairwise err (relative error) of columns' distribution.
        """
        # merge two frequency counts, and calculate relative difference
        df = pd.DataFrame(columns=self._columns, index=['err'])
        df.fillna(0)
        for col in self._columns:
            df.at['err', col] = relative_error(self.first[col].counts(),
                                               self.second[col].counts())
        return df

    def jsd(self):
        """
        Return pairwise JSD (Jensen-Shannon divergence) of columns'distribution.
        """
        df = pd.DataFrame(columns=self._columns, index=['jsd'])
        df.fillna(0)
        for col in self._columns:
            df.at['jsd', col] = jensen_shannon_divergence(
                self.first[col].counts(), self.second[col].counts())
        return df

    def corr(self):
        """
        Return pairwise correlation and dependence measured by mi (mutual
        information).
        """
        return self.first.mi(), self.second.mi()

    def dist(self, column):
        """
        Return frequency distribution of one column.

        Parameters
        ----------
        column : str
            column name, whose distribution will be return
        """
        if len(self._dt) == 0:
            for c in self._columns:
                self._dt[c] = {}
                if self.first[c].categorical:
                    bins = self.first[c].domain
                    counts1 = self.first[c].counts(bins=bins)
                    counts2 = self.second[c].counts(bins=bins)
                else:
                    min_, max_ = self.first[c].domain
                    # the domain from two data set are same;
                    # extend the domain to human-readable range
                    bins = normalize_range(min_, max_ + 1)
                    counts1 = self.first[c].counts(bins=bins)
                    counts2 = self.second[c].counts(bins=bins)
                    # Note: index, value of np.histogram has different length
                    bins = bins[:-1]
                self._dt[c]['bins'] = bins
                # stack arrays vertically
                self._dt[c]['counts'] = np.vstack((counts1, counts2))
        return self._dt[column]['bins'], self._dt[column]['counts']

    def describe(self):
        """
        Give descriptive difference between two data sets, which concluded
        relative errors, and jsd divergence.
        Return a panda.DataFrame, whose columns are two dataset's columns, and
        indexes are a array of metrics, e.g. ['err', 'jsd'].
        """
        df1 = self.err()
        df2 = self.jsd()
        return pd.concat([df1, df2])

    def classify(self, label: str, test: pd.DataFrame = None):
        """
        Train two svm classifiers based on data sets, and predict class labels
        for test data. Return both error rates.

        Parameters
        ----------
        label : str
            classifier feature, key is one column in left data frame.
            It supports two-class and multi-class.

        test : {pandas.DataFrame}
            test frame, is test data for machine learning algorithms. If it is
            not provided, it will split 20% of left data frame as test data.

        Returns
        -------
        a DataFrame, e.g.
                         target                         source     target
                      male female                    male female male female
        source male   1    3        or actual male   1    3      1    2
               female 2    4                  female 2    4      3    4
        """
        if (not self.first[label].categorical or
                not self.second[label].categorical):
            raise ValueError(
                f'Classifier can\'t run on non-categorical column: {label}')
        from sklearn.metrics import confusion_matrix

        def split_feature_label(df: pd.DataFrame):
            # TODO need improve sub_cols
            sub_cols = [attr for attr in df.columns if attr.startswith(label)]
            if len(sub_cols) == 0:
                return df, None
            is_one_class = len(sub_cols) == 2
            if is_one_class:
                # For one class, there are two sorted values.
                # e.g. ['Yes', 'No'] => [[0, 1],
                #                        [1, 0]]
                # Here it should choose second column to represent this attribute.
                label_ = sub_cols[1]
                return df.drop(sub_cols, axis=1), df[label_]
            else:
                try:
                    # merge multiple column into one: [Name_A, Name_B, ..] => Name
                    _y = df[sub_cols].apply(lambda x: Index(x).get_loc(1),
                                            axis=1)
                    return df.drop(sub_cols, axis=1), _y
                except KeyError as e:
                    print(e)
                    print(sub_cols)
                    print(df[sub_cols])

        # If test dataset is not provided, then split 20% of source dataset for testing.
        if test is None:
            lt, test = train_test_split(self.first, test_size=0.2)
            rt, _ = train_test_split(self.second, test_size=0.2)
        else:
            lt = self.first
            rt = self.second
        ts = self.first.encode(data=lt)
        lt_x, lt_y = split_feature_label(ts)
        test_x, test_y = split_feature_label(self.first.encode(data=test))
        rt_x, rt_y = split_feature_label(self.first.encode(data=rt))

        # construct svm classifier, and predict
        lt_yp = train_and_predict(lt_x, lt_y, test_x)
        rt_yp = train_and_predict(rt_x, rt_y, test_x)

        columns = self.first[label].bins
        labels = range(len(columns))
        # If test dataset has class label, return two expected score.
        if label in test:
            lcm = confusion_matrix(test_y, lt_yp, labels=labels)
            rcm = confusion_matrix(test_y, rt_yp, labels=labels)
            # lcm = lcm.astype('float') / lcm.sum(axis=1)
            # rcm = rcm.astype('float') / rcm.sum(axis=1)
            return (pd.DataFrame(lcm, columns=columns, index=columns),
                    pd.DataFrame(rcm, columns=columns, index=columns))
        # If test dataset don't have class label, return their expected values.
        else:
            cm = confusion_matrix(lt_yp, rt_yp, labels=labels)
            return pd.DataFrame(cm, columns=columns, index=columns)

    def to_html(self, buf=None, title='Evaluation Report', info=True,
                distribute=True, correlate=True, classifier=None, labels=None,
                test=None):
        """
        Render the evaluation result of two data set as an HTML file.

        Parameters
        ----------
        buf : optional
            buffer to write to

        title : str
            title of evaluation report

        info : bool, default true
            show basic information of two data set, including relative error,
            and Jensen-Shannon divergence (jsd).

        distribute : bool, default true
            show distribution of each attribute.

        correlate : bool, default true
            show correlation of pair-wise attributes.

        classifier : str
            use classifier to train data set on one or more columns (defined by
            parameter 'label') and show prediction result on the evaluation
            report. Optional classifier: SVM.

        labels : list of column names
            column name, or a list of column names separated by comma, used for
            classification task.

        test : pd.DataFrame
            test data for classification, and other machine learning tasks.
        """
        from ds4mllib.utils import (plot_histogram, plot_heatmap,
                                 plot_confusion_matrix)
        from mako.template import Template
        import os
        old_cwd = os.getcwd()
        os.chdir(os.path.dirname(__file__))
        template = Template(filename='template/report.html')
        os.chdir(old_cwd)

        topics = []
        content = {}
        # format different kinds of evaluation result to unified style
        if info:
            topics.append('basic')
            content['basic'] = [self.describe().to_dict('split')]

        if distribute:
            topics.append('dist')
            content['dist'] = []
            for col in self.columns:
                bins, counts = self.dist(col)
                svg = plot_histogram(bins, counts)
                content['dist'].append({'name': col, 'columns': bins,
                                        'data': counts, 'path': svg})
        if correlate:
            topics.append('corr')
            content['corr'] = []
            source_mi, target_mi = self.corr()
            source_svg = plot_heatmap(source_mi)
            target_svg = plot_heatmap(target_mi)
            content['corr'].append({'matrix': source_mi.to_dict('split'),
                                    'path': source_svg})
            content['corr'].append({'matrix': target_mi.to_dict('split'),
                                    'path': target_svg})

        if labels is not None:
            topics.append('svm')
            content['svm'] = []
            for col in labels:
                in_test = (test is not None and col in test) or (test is None)
                if in_test:
                    # When class label in svm classify test data, try to match
                    # two predicted result with the actual data, and so, there
                    # will be two confusion matrix diagrams.
                    try:
                        source_cm, target_cm = self.classify(col, test=test)
                        vrange = (
                            min(source_cm.values.min(), target_cm.values.min()),
                            max(source_cm.values.max(), target_cm.values.max()))
                        path = (
                            plot_confusion_matrix(source_cm, vrange=vrange,
                                                  xlabel='raw',
                                                  ylabel='actual'),
                            plot_confusion_matrix(target_cm, vrange=vrange,
                                                  xlabel='synth',
                                                  ylabel='actual'))
                        content['svm'].append({'column': col, 'path': path})
                    except ValueError as e:
                        print(e)
                else:
                    # If not, will compare two predicted result.
                    try:
                        cm = self.classify(col, test=test)
                        # make path's type: 1-tuple
                        path = (plot_confusion_matrix(cm, xlabel='synth',
                                                      ylabel='raw'),)
                        content['svm'].append({'column': col, 'path': path})
                    except ValueError as e:
                        print(e)

        svms = content['svm'] if 'svm' in content else []
        if buf:
            with open(buf, 'w+', encoding='utf-8') as file:
                file.write(template.render(title=title,
                                           basics=content['basic'],
                                           dists=content['dist'],
                                           corrs=content['corr'],
                                           svms=svms))



"""
DataSet: data structure for potentially mixed-type Attribute.
"""

from pandas import DataFrame

from ds4mllib.attribute import Attribute


class DataSet(DataFrame):

    _metadata = ['_categories']

    def __init__(self, data=None, index=None, columns=None, dtype=None,
                 copy=False, categories=None):
        """
        A DataFrame with categories information.

        Parameters
        ----------
        categories : list of columns
            Column names whose values are categorical.
        """
        DataFrame.__init__(self, data=data, index=index, columns=columns,
                           dtype=dtype, copy=copy)
        self.separator = '_'
        self._categories = categories or []

    @property
    def _constructor(self):
        return DataSet

    @property
    def _constructor_sliced(self):
        return Attribute

    # workaround: override method to add parameter 'categorical' for Attribute
    # constructor
    def _box_col_values(self, values, items):
        """
        Provide boxed values for a column.
        """
        klass = self._constructor_sliced
        return klass(values, index=self.index, name=items, fastpath=True,
                     categorical=items in self._categories)

    def mi(self):
        """ Return mutual information of pairwise attributes. """
        from ds4mllib.metrics import pairwise_mutual_information
        return pairwise_mutual_information(self)

    def encode(self, data=None):
        """
        Transform data set to values by kinds of encoders.
        If data is set, use this data set's encoders to transform.
        """
        # If the data to encode is None, then transform source data _data;
        frame = DataFrame()
        for col, attr in self.items():
            # TODO: Only focus on Bayesian Network nodes ??
            if data is not None and col not in data:
                continue
            if attr.atype == 'string' and attr.categorical:
                subs = attr.encode(None if data is None else data[col])
                for label in attr.bins:
                    frame[col + self.separator + label] = subs[label]
            elif attr.atype != 'string':
                frame[col] = attr.encode(None if data is None else data[col])
        return frame

    def _sampling_dataset(self, network, cond_prs, n):
        from numpy import random
        root_col = network[0][1][0]
        root_prs = cond_prs[root_col]

        columns = [root_col]  # columns from bayesian network
        for node, _ in network:
            columns.append(node)

        frame = DataFrame(columns=columns)  # encoded DataFrame
        frame[root_col] = random.choice(len(root_prs), size=n, p=root_prs)

        for child, parents in network:
            child_cond_prs = cond_prs[child]
            for indexes in child_cond_prs.keys():
                prs = child_cond_prs[indexes]
                indexes = list(eval(indexes))

                filters = ''
                for parent, value in zip(parents, indexes):
                    filters += f"(frame['{parent}']=={value})&"
                filters = eval(filters[:-1])
                size = frame[filters].shape[0]
                if size:
                    frame.loc[filters, child] = random.choice(len(prs),
                                                              size=size,
                                                              p=prs)
            child_prs = self[child].prs
            frame.loc[frame[child].isnull(), child] = random.choice(
                len(child_prs), size=frame[child].isnull().sum(), p=child_prs)
        frame[frame.columns] = frame[frame.columns].astype(int)
        return frame

    def synthesize(self, epsilon=0.1, degree=2,
                   pseudonyms=None, deletes=None, retains=None):
        """
        Synthesize data set by a bayesian network to infer attributes'
        dependence relationship and differential privacy to keep differentially
        private.
        """
        deletes = deletes or []
        pseudonyms = pseudonyms or []
        retains = retains or []

        columns = [col for col in self.columns if col not in deletes]
        nodes = set()  # nodes for bayesian networks
        for col in columns:
            if col in pseudonyms or (
                    self[col].atype == 'string' and not self[col].categorical):
                continue
            nodes.add(col)
        # main steps of private bayesian network for synthesis
        # encode dataset into bin indexes for bayesian network
        indexes = DataFrame()
        for col in nodes:
            indexes[col] = self[col].bin_indexes()
        if indexes.shape[1] < 2:
            raise Exception('If infer bayesian network, it requires at least 2 '
                            'attributes in dataset.')

        # Bayesian network is defined as a set of AP (attribute-parent) pairs.
        # e.g. [(x1, p1), (x2, p2), ...], and pi is the parents of xi.
        #
        # The algorithm follows the composability property of differential
        # privacy, so the privacy budget is split to two parts.
        from ds4mllib.synthesizer import greedy_bayes, noisy_conditionals
        network = greedy_bayes(indexes, epsilon / 2, degree=degree,
                               retains=retains)
        cond_prs = noisy_conditionals(network, indexes, epsilon / 2)

        sampling = self._sampling_dataset(network, cond_prs, self.shape[0])
        frame = DataFrame(columns=columns)
        for col, attr in self.items():
            if col in deletes:
                continue
            if col in pseudonyms:
                frame[col] = attr.pseudonymize()
                continue
            if col in retains:
                frame[col] = attr
                continue
            if col in sampling:
                frame[col] = attr.choice(indexes=sampling[col])
            elif not attr.categorical:
                frame[col] = attr.random()
            else:
                frame[col] = attr.choice()
        return frame



#!/usr/bin/env python

import setuptools


INSTALL_REQUIRES = [
    'numpy >= 1.14.3',
    'matplotlib >= 2.2.2',
    'mako ==1.0.12',
    'pandas >= 0.24.2',
    'scikit-learn >= 0.20.2',
    'pytest >= 4.6.2',
    'python-dateutil >= 2.7.3',
    'setuptools >= 39.1.0'
]



setuptools.setup(
    name='ds4mllib',
    description='A python library for data synthesis and evaluation',
    version='0.0.1',
    packages=setuptools.find_packages(),
    author=['Rongjun', 'Yan', 'David'],
    author_email= 'yan.zhao01@sap.com',
    install_requires=INSTALL_REQUIRES,
    platform='any'
)




"""
Attribute: data structure for 1-dimensional cross-sectional data
"""

import numpy as np

from bisect import bisect_right
from random import uniform
from pandas import Series, DataFrame
from dateutil.parser import parse
from datetime import datetime, timedelta

import ds4mllib
from ds4mllib import utils


class Attribute(Series):

    _epoch = datetime(1970, 1, 1)  # for datetime handling

    def __init__(self, data, name=None, dtype=None, index=None, copy=False,
                 fastpath=False, categorical=False):
        """
        A Series with extra information, e.g. categorical.

        Parameters
        ----------
        categorical : bool
            set categorical label for attribute. If categorical, this attribute
            takes on a limited and fixed number of possible values. Examples:
            blood type, gender.
        """
        Series.__init__(self, data, name=name, dtype=dtype, index=index,
                        copy=copy, fastpath=fastpath)

        # bins can be int (size of histogram bins), str (as algorithm name),
        self._bins = ds4mllib.params['attribute.bins']

        self._min = None
        self._max = None
        self._step = None

        # probability distribution (pr)
        self.bins = None
        self.prs = None

        from pandas.api.types import infer_dtype
        # atype: date type for handle different kinds of attributes in data
        # synthesis, support: integer, float, string, datetime.
        self.atype = infer_dtype(self, skipna=True)
        if self.atype == 'integer':
            pass
        elif self.atype == 'floating' or self.atype == 'mixed-integer-float':
            self.atype = 'float'
        elif self.atype in ['string', 'mixed-integer', 'mixed']:
            self.atype = 'string'
            if all(map(utils.is_datetime, self._values)):
                self.atype = 'datetime'

        # fill the missing values with the most frequent value
        self.fillna(self.mode()[0], inplace=True)

        # special handling for datetime attribute
        if self.atype == 'datetime':
            self.update(self.map(self._to_seconds).map(self._date_formatter))

        # how to define the attribute is categorical.
        self.categorical = categorical or (
                self.atype == 'string' and not self.is_unique)
        self._set_domain()
        self._set_distribution()

    # handling functions for datetime attribute
    def _to_seconds(self, timestr):
        return int((parse(timestr) - self._epoch).total_seconds())

    def _date_formatter(self, seconds):
        date = self._epoch + timedelta(seconds=seconds)
        return '%d/%d/%d' % (date.month, date.day, date.year)

    # Take pandas.Series as manipulation result.
    @property
    def _constructor(self):
        return Series

    @property
    def _constructor_expanddim(self):
        from ds4mllib.dataset import DataSet
        return DataSet

    @property
    def is_numerical(self):
        return self.atype == 'integer' or self.atype == 'float'

    @property
    def domain(self):
        """
        Return attribute's domain, which can be a list of values for categorical
        attribute, and an interval with min/max value for non-categorical
        attribute.
        """
        if self.categorical:
            return self.bins
        else:
            return [self._min, self._max]

    @domain.setter
    def domain(self, domain: list):
        """
        Set attribute's domain, includes min, max, frequency, or distribution.

        Generally, the domain of one attribute can be calculated automatically.
        This method can be manually called for specific purposes, e.g. compare
        two same attributes based on same domain.

        Parameters
        ----------
        domain : list
            domain of one attribute. For numerical or datetime attributes, it
            should be a list of two elements [min, max]; For categorical
            attributes, it should a list of potential values of this attribute.
        """
        # if a attribute is numerical and categorical and domain's length is
        # bigger than 2, take it as categorical. e.g. zip code.
        if self.atype == 'datetime':
            domain = list(map(self._to_seconds, domain))
        if (self.is_numerical and self.categorical and len(domain) > 2) or (
                self.categorical):
            self._min = min(domain)
            self._max = max(domain)
            self.bins = np.array(domain)
        elif self.is_numerical:
            self._min, self._max = domain
            self._step = (self._max - self._min) / self._bins
            self.bins = np.array([self._min, self._max])
        elif self.atype == 'string':
            lengths = [len(str(i)) for i in domain]
            self._min = min(lengths)
            self._max = max(lengths)
            self.bins = np.array(domain)
        self._set_distribution()

    def _set_domain(self):
        """
        Compute domain (min, max, distribution bins) from input data
        """
        if self.atype == 'string':
            self._items = self.astype(str).map(len)
            self._min = int(self._items.min())
            self._max = int(self._items.max())
            if self.categorical:
                self.bins = self.unique()
            else:
                self.bins = np.array([self._min, self._max])
        elif self.atype == 'datetime':
            self.update(self.map(self._to_seconds))
            if self.categorical:
                self.bins = self.unique()
            else:
                self._min = float(self.min())
                self._max = float(self.max())
                self.bins = np.array([self._min, self._max])
                self._step = (self._max - self._min) / self._bins
        else:
            self._min = float(self.min())
            self._max = float(self.max())
            if self.categorical:
                self.bins = self.unique()
            else:
                self.bins = np.array([self._min, self._max])
                self._step = (self._max - self._min) / self._bins

    def _set_distribution(self):
        if self.categorical:
            counts = self.value_counts()
            for value in set(self.bins) - set(counts.index):
                counts[value] = 0
            counts.sort_index(inplace=True)
            if self.atype == 'datetime':
                counts.index = list(map(self._date_formatter, counts.index))
            self._counts = counts.values
            self.prs = utils.normalize_distribution(counts)
            self.bins = np.array(counts.index)
        else:
            # Note: hist, edges = numpy.histogram(), all but the last bin
            # is half-open. If bins is 20, then len(hist)=20, len(edges)=21
            if self.atype == 'string':
                hist, edges = np.histogram(self._items,
                                           bins=self._bins)
            else:
                hist, edges = np.histogram(self, bins=self._bins,
                                           range=(self._min, self._max))
            self.bins = edges[:-1]  # Remove the last bin edge
            self._counts = hist
            self.prs = utils.normalize_distribution(hist)
            if self.atype == 'integer':
                self._min = int(self._min)
                self._max = int(self._max)

    def counts(self, bins=None):
        """
        Return an array containing counts of unique values.

        This function works with `attribute.bins`. Combination of both are
        like `Series.value_counts`. The parameter `bins` can be none, or a list.
        """
        if bins is None:
            return self._counts
        if self.categorical:
            if self.atype == 'datetime':
                bins = list(map(self._to_seconds, bins))
            counts = self.value_counts()
            for value in set(bins) - set(counts.index):
                counts[value] = 0
            return np.array([counts.get(b) for b in bins])
        else:
            if len(bins) == 1:
                return np.array([self.size])
            hist, _ = np.histogram(self, bins=bins)
            return hist

    def bin_indexes(self):
        """
        Encode values into bin indexes for Bayesian Network.
        """
        if self.categorical:
            mapping = {value: idx for idx, value in enumerate(self.bins)}
            indexes = self.map(lambda x: mapping[x], na_action='ignore')
        else:
            indexes = self.map(lambda x: bisect_right(self.bins, x) - 1,
                               na_action='ignore')
        indexes.fillna(len(self.bins), inplace=True)
        return indexes.astype(int, copy=False)

    def metadata(self):
        """
        Return attribution's metadata information in JSON format or Python
        dictionary. Usually used in debug and testing.
        """
        return {
            'name': self.name,
            'dtype': self.dtype,
            'atype': self.atype,
            'categorical': self.categorical,
            'min': self._min,
            'max': self._max,
            'bins': self.bins.tolist(),
            'prs': self.prs.tolist()
        }

    def pseudonymize(self):
        """
        Return pseudonymized values for this attribute, which is used to
        substitute identifiable data with a reversible, consistent value.
        """
        if self.categorical:
            mapping = {b: utils.pseudonymise_string(b) for b in self.bins}
            return self.map(lambda x: mapping[x])

        if self.atype == 'string':
            return self.map(utils.pseudonymise_string)
        elif self.is_numerical or self.atype == 'datetime':
            return self.map(str).map(utils.pseudonymise_string)

    def random(self, size=None):
        """
        Return an random array with same length (usually used for
        non-categorical attribute).
        """
        if size is None:
            size = len(self)
        if self._min == self._max:
            rands = np.ones(size) * self._min
        else:
            rands = np.arange(self._min, self._max,
                              (self._max - self._min) / size)

        np.random.shuffle(rands)
        if self.atype == 'string':
            if self._min == self._max:
                length = self._min
            else:
                length = np.random.randint(self._min, self._max)
            vectorized = np.vectorize(lambda x: utils.randomize_string(length))
            rands = vectorized(rands)
        elif self.atype == 'integer':
            rands = list(map(int, rands))
        elif self.atype == 'datetime':
            rands = list(map(self._date_formatter, rands))
        return Series(rands)

    def _sampling_bins(self, index: int):
        if self.categorical:
            return self.bins[index]

        length = len(self.bins)
        if index < length - 1:
            return uniform(self.bins[index], self.bins[index + 1])
        else:
            return uniform(self.bins[-1], self._max)

    def choice(self, size=None, indexes=None):
        """
        Return a random sample based on this attribute's probability.
        If indexes and n are both set, ignore n.

        Parameters
        ----------
        size : int
            size of random sample

        indexes : array-like
            array of indexes in bins
        """
        if indexes is None:
            if size is None:
                size = len(self)
            indexes = Series(np.random.choice(len(self.prs),
                                              size=size, p=self.prs))
        column = indexes.map(lambda x: self._sampling_bins(x))
        if self.atype == 'datetime':
            if not self.categorical:
                column = column.map(self._date_formatter)
        elif self.atype == 'integer':
            column = column.round().astype(int)
        elif self.atype == 'string':
            if not self.categorical:
                column = column.map(lambda x: utils.randomize_string(int(x)))
        return column

    def encode(self, data=None):
        """
        Encode labels to normalized encoding.

        Parameters
        ----------
        data : array-like
            target values
        """
        if data is None:
            data = self.copy()
        else:
            if self.atype == 'datetime':
                if all(map(utils.is_datetime, data)):
                    data = data.map(self._to_seconds)
                else:
                    data = data.map(int)

        if self.categorical:
            df = DataFrame()
            for c in self.bins:
                df[c] = data.apply(lambda v: 1 if v == c else 0)
            return df

        if self.atype != 'string':
            return data.apply(lambda v:  # 1e-8 is a small delta
                              int((v - self._min) / (self._step + 1e-8))
                              / self._bins)
        else:
            raise ValueError('Non-categorical attribute does not support encode '
                             'method.')
